@@ -6,7 +6,7 @@
 import logging
 import os
 import re
-from collections import defaultdict
+from collections import defaultdict, Counter
 from io import StringIO
 from pprint import pprint
 
@@ -20,6 +20,7 @@ import simplejson as json
 import yaml
 
 ureg = pint.UnitRegistry()
+UNIQUENESS_SEP = " & "
 
 
 def yamlcat(*files):
@@ -295,6 +296,7 @@ class DataFrameSchematizer:
         self._is_units = False
         self._source_units = None
         self._target_units = None
+        self.uniqueness_sets = defaultdict(list)
 
     def build(self):
         """build and return schema"""
@@ -307,6 +309,13 @@ class DataFrameSchematizer:
         for colname, desc in self.columns_specs.items():
             schema["properties"][colname] = desc
         schema["required"] = self.required
+        for uniqueness_tag, columns in self.uniqueness_sets.items():
+            dummy_col = UNIQUENESS_SEP.join(columns)
+            schema["properties"][dummy_col] = {
+                "items": {"type": "string"},
+                "type": "array",
+                "uniqueItems": True,
+            }
         return schema
 
     def _add_columns_from_json(self, jsontxt):
@@ -339,6 +348,7 @@ class DataFrameSchematizer:
         types=("integer",),
         unique=False,
         mandatory=False,
+        uniqueness_id=None,
         units=None,
         **kwargs,
     ):
@@ -350,6 +360,8 @@ class DataFrameSchematizer:
             self.required.append(name)
         else:
             types.append("null")
+        if uniqueness_id:
+            self.uniqueness_sets[uniqueness_id].append(name)
         # ---------------------------------------------------------
         if len(types) > 1:
             items = {"anyOf": [{"type": typ} for typ in types]}
@@ -444,6 +456,16 @@ class DataFrameSchematizer:
         if fillnas:
             df = df.fillna(value=fillnas, downcast="infer")
         # =====================================================================
+        # build dummy columns if required by the multiple-columns uniqueness
+        # =====================================================================
+        dummies = []
+        for uniqueness_tag, columns in self.uniqueness_sets.items():
+            _df = df[columns].astype(str)
+            dummy_col = UNIQUENESS_SEP.join(columns)
+            dummies.append(dummy_col)  # to delete them later on
+            df[dummy_col] =  _df.apply(UNIQUENESS_SEP.join, axis=1)
+
+        # =====================================================================
         # second validation
         # =====================================================================
         # df -> dict -> json -> dict to convert NaN to None
@@ -453,6 +475,8 @@ class DataFrameSchematizer:
             check_types=False,
             # report=early_report,
         )
+        if dummies:
+            df.drop(columns=dummies, inplace=True)
         report = {**early_report, **report}
         return df, len(report) == 0, dict(report)
 
@@ -487,13 +511,23 @@ def _validate(document, validator, check_types=None, report=None):
             # want typing-related issues
             continue
         try:
+            # generic regular validation catcher
             col, row, *rows = error.absolute_path
         except ValueError:
-            report["general"].append(error.message)
-        else:
-            if error.message not in report[(col, row)]:
-                report[(col, row)].append(error.message)
-                report[(col, row)].extend([e.message for e in error.context])
+            if error.absolute_schema_path[-1] == "uniqueItems":
+                non_uniques = {k for k, v in Counter(error.instance).items() if v>1}
+                col = error.absolute_schema_path[-2]
+                row = ("?")
+                if len(non_uniques) > 1:
+                    error.message = "values %s are not unique" % non_uniques
+                else:
+                    error.message = "value %s is not unique" % non_uniques
+            else:
+                report["general"].append(error.message)
+                continue
+        if error.message not in report[(col, row)]:
+            report[(col, row)].append(error.message)
+            report[(col, row)].extend([e.message for e in error.context])
     return report
 
 
